/*
 * Copyright (C) 2009 Jeffrey Knockel
 * 
 * This file is a part of MIPS StackSmash.
 * 
 * MIPS StackSmash is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 * 
 * MIPS StackSmash is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with MIPS StackSmash.  If not, see <http://www.gnu.org/licenses/>.
 */

#include <stdlib.h>
#include <stdio.h>
#include <stdint.h>
#include <string.h>
#include <unistd.h>
#include <errno.h>
#include <sys/ptrace.h>

//#define INSTR 0x03a00008 /* jr sp */
#define INSTR 0x00800008 /* jr a0 */
//#define INSTR 0x00a00008 /* jr a1 */
//#define INSTR 0x00400008 /* jr v0 */
//#define INSTR 0x01000008 /* jr t0 */

int main( int argc, char **argv )
{
    pid_t pid;
    uint32_t *addr1, *addr2;
    uint32_t instr;
    char *ret;
    char buffer[81];
    
    if (argc != 2) {
        fprintf (stderr, "Usage: %s PID\n", argv[0]);
        exit (EXIT_FAILURE);
    }

    pid = (pid_t) strtoll (argv[1], NULL, 10);

    if (ptrace (PTRACE_ATTACH, pid, NULL, NULL) != 0) {
        fprintf (stderr, "%s: ", argv[0]);
        perror ("ptrace");
        exit (EXIT_FAILURE);
    }

    for (;;) {
        errno = 0;
        ret = fgets (buffer, sizeof(buffer), stdin);
        if (ret == NULL) {
            if (errno != 0) {
                fprintf (stderr, "%s: ", argv[0]);
                perror ("fgets");
                exit (EXIT_FAILURE);
            }
            break;
        }
        if (strchr (buffer, '\n') == NULL) {
            fprintf (stderr, "%s: bad range: %s\n", argv[0], buffer);
            continue;
        }
        if (sscanf (buffer, "%p-%p\n", &addr1, &addr2) != 2) {
            fprintf (stderr, "%s: bad range: %s", argv[0], buffer);
            continue;
        }
        fprintf (stderr, "%s: examining %p-%p\n", argv[0], addr1, addr2);
        for (; addr1 < addr2; ++addr1) {
            instr = (uint32_t) ptrace (PTRACE_PEEKDATA, pid, addr1, NULL);
            if (errno != 0) {
                fprintf (stderr, "%s: ", argv[0]);
                perror ("ptrace");
                if (ptrace (PTRACE_DETACH, pid, NULL, NULL) != 0) {
                    fprintf (stderr, "%s: ", argv[0]);
                    perror ("ptrace");
                }
                exit (EXIT_FAILURE);
            }
            if (instr == INSTR) {
                printf ("%p\n", addr1);
            }
        }
    }
    if (ptrace (PTRACE_DETACH, pid, NULL, NULL) != 0) {
        fprintf (stderr, "%s: ", argv[0]);
        perror ("ptrace");
        exit (EXIT_FAILURE);
    }

    return 0;
}
