/*
 * Copyright (C) 2009 Jeffrey Knockel
 * 
 * This file is a part of MIPS StackSmash.
 * 
 * MIPS StackSmash is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 * 
 * MIPS StackSmash is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with MIPS StackSmash.  If not, see <http://www.gnu.org/licenses/>.
 */

#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <limits.h>
#include <unistd.h>
#include <sys/wait.h>
#include <sys/socket.h>
#include <netinet/in.h>

#define PORT 44444
#define BACKLOG 5

static void handle_sigchld( int signal )
{
    (void) signal;
    while (waitpid (-1, NULL, WNOHANG) > 0);
}

static void echo( int sockfd )
{
    char buffer[80];
    int len;

    memset (buffer, 0, sizeof(buffer));

    recv (sockfd, buffer, 256, 0);

    len = strlen (buffer);

    if (len > 0) {
        send (sockfd, buffer, len, 0);
    }
}

int main( void )
{
    struct sigaction act;
    struct sockaddr_in sin;
    int sockfd;
    int newsockfd;
    socklen_t len;
    int val = 1;

    memset (&act, 0, sizeof(act));
    act.sa_handler = &handle_sigchld;
    act.sa_flags = SA_RESTART;

    if (sigaction (SIGCHLD, &act, 0)) {
        perror ("server: sigaction");
        exit (EXIT_FAILURE);
    }

    memset (&sin, 0, sizeof(sin));
    sin.sin_family = AF_INET;
    sin.sin_addr.s_addr = INADDR_ANY;
    sin.sin_port = htons (PORT);

    if ((sockfd = socket (PF_INET, SOCK_STREAM, 0)) < 0) {
        perror ("server: socket");
        exit (EXIT_FAILURE);
    }

    setsockopt (sockfd, SOL_SOCKET, SO_REUSEADDR, &val, sizeof(val));

    if (bind (sockfd, (struct sockaddr *) &sin, sizeof(sin)) < 0) {
        perror ("server: bind");
        exit (EXIT_FAILURE);
    }

    listen (sockfd, BACKLOG);

    for (;;) {
        len = sizeof(sin);
        if ((newsockfd = accept (sockfd, (struct sockaddr *) &sin, &len)) < 0) {
            perror ("server: accept");
            exit (EXIT_FAILURE);
        }
        if (fork () == 0) {
            fprintf (stderr, "server: accepted connection...\n");
            close (sockfd);
            echo (newsockfd);
            return 0;
        }
        close (newsockfd);
    }

    return 0;
}
